import h5py
import numpy
import os
import pytest
from tempfile import TemporaryDirectory

@pytest.fixture(scope='module')
def simple_h5_file():
    with TemporaryDirectory() as td:
        f = h5py.File(os.path.join(td, "example.h5"))
        f.create_dataset("/group1/subgroup1/dataset1", (200,), dtype='u8')
        f.create_dataset("/group1/subgroup1/dataset2", (2, 128, 500), dtype='f4')
        f.create_dataset("/group1/subgroup2/dataset1", (12,), dtype='i2')
        f['group1'].attrs['string'] = 'foo'
        f['group1'].attrs['array'] = numpy.zeros((3, 4))
        yield f



import io
from h5glance import terminal

def test_tree_view(simple_h5_file):
    sio = io.StringIO()
    tvb = terminal.TreeViewBuilder()
    tree = tvb.object_node(simple_h5_file, simple_h5_file.filename)
    terminal.print_tree(tree, file=sio)
    out = sio.getvalue()
    assert '└dataset2' in out
    assert 'uint64' in out
    assert '128 × 500' in out

def test_attributes(simple_h5_file):
    tvb = terminal.TreeViewBuilder()
    r1, c1 = tvb.object_node(simple_h5_file['group1'], 'group1')
    assert '2 attributes' in r1
    assert 'attributes' not in c1[0][0]

    tvb2 = terminal.TreeViewBuilder(expand_attrs=True)
    r2, c2 = tvb2.object_node(simple_h5_file['group1'], 'group1')
    assert 'attributes' not in r2
    assert '2 attributes' in c2[0][0]
    attr_lines = {node[0] for node in c2[0][1]}
    assert len(attr_lines) == 2


def test_dataset_info(simple_h5_file):
    sio = io.StringIO()
    terminal.print_dataset_info(simple_h5_file["/group1/subgroup1/dataset2"], file=sio)
    out = sio.getvalue()
    assert 'dtype: float32' in out
    assert 'shape: 2 × 128 × 500' in out

def test_completer(simple_h5_file):
    comp = terminal.H5Completer(simple_h5_file)
    # Complete groups
    assert set(comp.completions('group1/sub')) == {'group1/subgroup1/', 'group1/subgroup2/'}
    # Complete a dataset - no trailing slash
    assert set(comp.completions('group1/subgroup2/')) == {'group1/subgroup2/dataset1'}
    # Case insensitive completion
    assert set(comp.completions('GRO')) == {'group1/'}



"""Command line h5glance-html interface for writing and serving HTML views of HDF5
"""
import argparse
import h5py
from http.server import HTTPServer, BaseHTTPRequestHandler
from pathlib import Path
import sys
import threading
import webbrowser

from .generate import make_document

def main(argv=None):
    ap = argparse.ArgumentParser(prog="h5glance",
                                 description="View HDF5 file structure in HTML")
    ap.add_argument("input", help="HDF5 file to view", type=Path)
    ap.add_argument("-w", "--write", metavar="HTML_FILE",
                    help="Write output to HTML file.")
    args = ap.parse_args(argv)

    if not args.input.is_file():
        print("Not a file:", args.input)
        sys.exit(2)
    elif not h5py.is_hdf5(args.input):
        print("Not an HDF5 file:", args.input)
        sys.exit(2)

    if args.write:
        with open("test_output.html", 'w') as f:
            f.write(str(make_document(args.input)))
            return

    serve(args.input)

def serve(h5path):
    class H5ViewHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            if self.path != "/":
                return self.send_error(404)
            self.send_response(200)
            self.end_headers()
            self.wfile.write(str(make_document(h5path)).encode('utf-8'))

    server = HTTPServer(('localhost', 0), H5ViewHandler)
    url = "http://{}:{}/".format(server.server_name, server.server_port)
    print("Serving on", url)
    t = threading.Timer(0.5, webbrowser.open_new_tab, args=(url,))
    t.start()
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        pass



"""Terminal h5glance interface for inspecting HDF5 files
"""
import argparse
import h5py
import io
import os
import numpy
from pathlib import Path
import shlex
from shutil import get_terminal_size
from subprocess import run
import sys

def fmt_shape(shape):
    return " × ".join(('Unlimited' if n is None else str(n)) for n in shape)

layout_names = {
    h5py.h5d.COMPACT: 'Compact',
    h5py.h5d.CONTIGUOUS: 'Contiguous',
    h5py.h5d.CHUNKED: 'Chunked',
    h5py.h5d.VIRTUAL: 'Virtual',
}

def fmt_attr(v):
    if isinstance(v, numpy.ndarray):
        sv = 'array [{}: {}]'.format(v.dtype.name, fmt_shape(v.shape))
    else:
        sv = repr(v)
        if len(sv) > 50:
            sv = sv[:20] + '...' + sv[-20:]
    return sv

def print_dataset_info(ds: h5py.Dataset, file=None):
    """Print detailed information for an HDF5 dataset."""
    print('      dtype:', ds.dtype.name, file=file)
    print('      shape:', fmt_shape(ds.shape), file=file)
    print('   maxshape:', fmt_shape(ds.maxshape), file=file)
    layout = ds.id.get_create_plist().get_layout()
    print('     layout:', layout_names.get(layout, 'Unknown'), file=file)
    if layout == h5py.h5d.CHUNKED:
        print('      chunk:', fmt_shape(ds.chunks), file=file)
        print('compression: {} (options: {})'
              .format(ds.compression, ds.compression_opts), file=file)

    if sys.stdout.isatty():
        numpy.set_printoptions(linewidth=get_terminal_size()[0])

    print('\nsample data:', file=file)
    if ds.ndim == 0:
        print(ds[()], file=file)
    elif ds.ndim == 1:
        print(ds[:10], file=file)
    else:
        select = (0,) * (ds.ndim - 2) + (slice(0, 10),) * 2
        print(ds[select], file=file)

    print('\n{} attributes:'.format(len(ds.attrs)), file=file)
    for k, v in ds.attrs.items():
        print('* ', k, ': ', fmt_attr(v), sep='', file=file)

class ColorsNone:
    dataset = group = link = reset = ''

class ColorsDefault:
    dataset = '\x1b[1m'  # Bold
    group = '\x1b[94m'   # Bright blue
    link = '\x1b[95m'    # Bright magenta
    reset = '\x1b[0m'

def use_colors():
    if os.name != 'posix':
        return False
    env = os.environ.get('H5GLANCE_COLORS', '')
    if env:
        return env != '0'
    return sys.stdout.isatty()

class TreeViewBuilder:
    """Build a tree view of an HDF5 group or file

    The tree nodes are tuples (line, children).
    """
    def __init__(self, expand_attrs=False):
        self.expand_attrs = expand_attrs
        if use_colors():
            self.colors = ColorsDefault
        else:
            self.colors = ColorsNone

    def object_node(self, obj, name):
        """Build a tree node for an HDF5 group/dataset
        """
        children = []
        detail = attr_detail = ''
        color_start = color_stop = ''

        if self.expand_attrs:
            children += attrs_tree_nodes(obj)
        else:
            n = len(obj.attrs)
            if n:
                attr_detail = ' ({} attributes)'.format(n)

        if isinstance(obj, h5py.Dataset):
            color_start = self.colors.dataset
            color_stop = self.colors.reset
            detail = '\t[{dt}: {shape}]'.format(
                dt=obj.dtype.name,
                shape=fmt_shape(obj.shape),
            )
            if obj.id.get_create_plist().get_layout() == h5py.h5d.VIRTUAL:
                detail += ' virtual'
        elif isinstance(obj, h5py.Group):
            color_start = self.colors.group
            color_stop = self.colors.reset
            children += [self.group_item_node(obj, key)
                         for key in obj]
        else:
            detail = ' (unknown h5py type)'

        return (color_start + name + color_stop + detail + attr_detail), children

    def group_item_node(self, group, key):
        """Build a tree node for one key in a group"""
        link = group.get(key, getlink=True)
        if isinstance(link, h5py.SoftLink):
            target = link.path
        elif isinstance(link, h5py.ExternalLink):
            target = '{}/{}'.format(link.filename, link.path)
        else:
            return self.object_node(group[key], key)

        line = '{}{}{}\t-> {}'.format(
            self.colors.link, key, self.colors.reset, target)
        return line, []

def attrs_tree_nodes(obj):
    """Build tree nodes for attributes"""
    nattr = len(obj.attrs)
    if not nattr:
        return []

    children = [('{}: {}'.format(k, fmt_attr(v)), [])
                for (k, v) in obj.attrs.items()]
    return [('{} attributes:'.format(nattr), children)]

def print_tree(node, prefix1='', prefix2='', file=None):
    """Render a tree to show in the terminal.

    Each tree node consists of a line of text to be displayed
    and a list of child nodes.
    """
    root, children = node
    print(prefix1 + root, file=file)

    nchild = len(children)
    for i, node in enumerate(children):
        islast = (nchild == i + 1)
        c_prefix1 = prefix2 + ('└'  if islast else '├')
        c_prefix2 = prefix2 + ('  ' if islast else '│ ')
        print_tree(node, prefix1=c_prefix1, prefix2=c_prefix2, file=file)

def page(text):
    """Display text in a terminal pager

    Respects the PAGER environment variable if set.
    """
    pager_cmd = shlex.split(os.environ.get('PAGER') or 'less -r')
    run(pager_cmd, input=text.encode('utf-8'))

def display_h5_obj(file: h5py.File, path=None, expand_attrs=False):
    """Display information on an HDF5 file, group or dataset

    This is the central function for the h5glance command line tool.
    """
    sio = io.StringIO()
    if path:
        root = file.filename + '/' + path.lstrip('/')
        obj = file[path]
    else:
        root = file.filename
        obj = file

    if isinstance(obj, h5py.Group):
        tvb = TreeViewBuilder(expand_attrs=expand_attrs)
        print_tree(tvb.object_node(obj, root), file=sio)
    elif isinstance(obj, h5py.Dataset):
        print(root, file=sio)
        print_dataset_info(obj, file=sio)
    else:
        sys.exit("What is this? " + repr(obj))

    # If the output has more lines than the terminal, display it in a pager
    output = sio.getvalue()
    if sys.stdout.isatty():
        nlines = len(output.splitlines())
        _, term_lines = get_terminal_size()
        if nlines > term_lines:
            return page(output)

    print(output)

class H5Completer:
    """Readline tab completion for paths inside an HDF5 file"""
    def __init__(self, file: h5py.File):
        self.file = file
        self.cache = (None, [])

    def completions(self, text: str):
        if text == self.cache[0]:
            return self.cache[1]
        prev_path, _, prefix = text.rpartition('/')
        group = self.file
        if prev_path:
            group = self.file[prev_path]
            prev_path += '/'
        res = [prev_path + k + ('/' if isinstance(v, h5py.Group) else '')
               for (k, v) in group.items()
               if k.lower().startswith(prefix.lower())]
        self.cache = (text, res)
        return res

    def rlcomplete(self, text: str, state: int):
        # print(repr(text), state)
        try:
            res = self.completions(text)
        except Exception as e:
            #print(e)
            return None
        # print(repr(text), len(res))
        if state >= len(res):
            return None
        return res[state]

def prompt_for_path(filename):
    """Prompt the user for a path inside the HDF5 file"""
    import readline
    with h5py.File(filename, 'r') as f:
        compl = H5Completer(f)
        readline.set_completer(compl.rlcomplete)
        readline.set_completer_delims('')
        readline.parse_and_bind("tab: complete")
        while True:
            res = input("Object path: {}/".format(filename))
            if f.get(res) is not None:
                print()
                return res
            print("No object at", repr(res))


def main(argv=None):
    from . import __version__
    ap = argparse.ArgumentParser(prog="h5glance",
             description="View HDF5 file structure in the terminal")
    ap.add_argument("file", help="HDF5 file to view", type=Path)
    ap.add_argument("path", nargs='?',
        help="Object to show within the file, or '-' to prompt for a name"
    )
    ap.add_argument('--attrs', action='store_true',
        help="Show attributes of groups",
    )
    ap.add_argument('--version', action='version',
                    version='H5glance {}'.format(__version__))

    args = ap.parse_args(argv)

    if not args.file.is_file():
        print("Not a file:", args.file)
        sys.exit(2)
    elif not h5py.is_hdf5(args.file):
        print("Not an HDF5 file:", args.file)
        sys.exit(2)

    path = args.path
    if path == '-':
        path = prompt_for_path(args.file)

    with h5py.File(args.file, 'r') as f:
        display_h5_obj(f, path, expand_attrs=args.attrs)



import h5py
from htmlgen import (Document, Element, Division, UnorderedList, Checkbox,
                     Label, ListItem, html_attribute, Span, Link, Script,
                     )
from pathlib import Path

_PKGDIR = Path(__file__).parent

class Style(Element):
    def __init__(self, styles):
        super().__init__('style')
        self.children.append_raw(styles)

class Abbreviation(Element):
    def __init__(self, *content, title=""):
        super().__init__("abbr")
        self.extend(content)
        self.title = title

    title = html_attribute('title')

class Code(Element):
    """An HTML in-line <code> element.
    """
    def __init__(self, *content):
        super().__init__("code")
        self.extend(content)

def id_generator(template):
    n = 0
    while True:
        yield template % n
        n += 1

dtype_kind_to_descr = {
    "u": "unsigned integer",
    "i": "signed integer",
    "f": "floating point",
    "c": "complex floating point",
    "b": "boolean",
    "O": "object (e.g. string)",
}

def make_dtype_abbr(dtobj):
    desc = ""
    if dtobj.kind in "uifc":
        desc += "{}-bit ".format(dtobj.itemsize * 8)
    try:
        desc += dtype_kind_to_descr[dtobj.kind]
    except KeyError:
        pass
    code = Code(dtobj.str)
    if desc:
        return Abbreviation(code, title=desc)
    return code

def make_list(*items):
    ul = UnorderedList()
    ul.extend(items)
    return ul

checkbox_ids = id_generator("h5glance-expand-switch-%d")

def checkbox_w_label(label_content):
    c = Checkbox()
    c.id = next(checkbox_ids)
    l = Label(label_content)
    l.for_ = c.id
    return [c, l]

def item_for_dataset(name, ds):
    shape = " × ".join(str(n) for n in ds.shape)
    namespan = Span(name)
    namespan.add_css_classes("h5glance-dataset-name")
    copylink = Link("#", "[📋]")
    copylink.set_attribute("data-hdf5-path", ds.name)
    copylink.add_css_classes("h5glance-dataset-copylink")
    li = ListItem(
        namespan,  " ", copylink, ": ",
        shape, " entries, dtype: ", make_dtype_abbr(ds.dtype)
    )
    li.add_css_classes("h5glance-dataset")
    return li

def item_for_group(gname, grp):
    subgroups, datasets = [], []
    for name, obj in sorted(grp.items()):
        if isinstance(obj, h5py.Group):
            subgroups.append((name, obj))
        else:
            datasets.append((name, obj))

    return ListItem(*checkbox_w_label(gname), make_list(
        *[item_for_group(n, g) for n, g in subgroups],
        *[item_for_dataset(n, d) for n, d in datasets],
    ))

def file_or_grp_name(obj):
    if isinstance(obj, h5py.File):
        return obj.filename
    elif isinstance(obj, h5py.Group):
        return obj.name
    return obj

treeview_ids = id_generator("h5glance-container-%d")

def make_fragment(obj):
    if isinstance(obj, h5py.Group):
        name = file_or_grp_name(obj)
        ct = make_list(item_for_group(name, obj))
    elif isinstance(obj, (str, Path)) and h5py.is_hdf5(obj):
        with h5py.File(obj, 'r') as f:
            return make_fragment(f)
    else:
        raise TypeError("Unknown object type: {!r}".format(obj))

    # Expand first level
    first_chkbx = ct.children.children[0].children.children[0] # Yuck
    first_chkbx.checked = True

    tv = Division(ct)
    tv.add_css_classes("h5glance-css-treeview")
    tv.id = next(treeview_ids)
    return tv

def get_treeview_css():
    with (_PKGDIR / 'treeview.css').open() as f:
        return Style(f.read())

JS_ACTIVATE_COPYLINKS_DOC = """
window.addEventListener("load", function(event) {
  enable_copylinks(document);
});
"""

JS_ACTIVATE_COPYLINKS_FRAG = """
enable_copylinks(document.getElementById("TREEVIEW-ID"));
"""

def get_copylinks_js(activation):
    with (_PKGDIR / "copypath.js").open() as f:
        return f.read().replace("//ACTIVATE", activation)

def make_document(obj):
    d = Document()
    d.append_head(get_treeview_css())
    d.append_head(Script(script=get_copylinks_js(JS_ACTIVATE_COPYLINKS_DOC)))
    d.title = "{} - H5Glance".format(file_or_grp_name(obj))
    d.append_body(make_fragment(obj))
    return d

def h5obj_to_html(obj):
    treeview = make_fragment(obj)
    js_activate = JS_ACTIVATE_COPYLINKS_FRAG.replace("TREEVIEW-ID", treeview.id)

    div = Division(
        get_treeview_css(),
        treeview,
        Script(script=get_copylinks_js(js_activate)),
    )
    return str(div)

class H5Glance:
    """View an HDF5 object in a Jupyter notebook"""
    def __init__(self, obj):
        self.obj = obj

    def _repr_html_(self):
        return h5obj_to_html(self.obj)

def install_ipython_h5py_display():
    """Call inside IPython to install HTML views for h5py groups and files"""
    from IPython import get_ipython
    ip = get_ipython()
    if ip is None:
        raise EnvironmentError("This function is to be called in IPython")
    html_formatter = ip.display_formatter.formatters['text/html']
    html_formatter.for_type(h5py.Group, h5obj_to_html)



from h5glance import generate


def test_h5obj_to_html(simple_h5_file):
    h = generate.h5obj_to_html(simple_h5_file)
    assert 'subgroup1' in h
    assert '</div>' in h
    assert '<!DOCTYPE' not in h  # This is not a full document

def test_make_document(simple_h5_file):
    h = str(generate.make_document(simple_h5_file))
    assert 'subgroup1' in h
    assert '<!DOCTYPE' in h



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['h5glance']

package_data = \
{'': ['*']}

install_requires = \
['h5py', 'htmlgen']

entry_points = \
{'console_scripts': ['h5glance = h5glance.terminal:main',
                     'h5glance-html = h5glance.html_cli:main']}

setup(name='h5glance',
      version='0.3',
      description='Explore HDF5 files in an HTML view',
      author='Thomas Kluyver',
      author_email='thomas@kluyver.me.uk',
      url='https://github.com/takluyver/h5glance',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.5',
     )
